import { BreakableBlockPickupCallbacks } from "./pickups.ts";
import {Circle, Point} from "./engine/datatypes.ts";
import { displace } from "./physics.ts";
import { getHuntMode } from "./huntmode.ts";
import { DrawPile } from "./drawpile.ts";
import { FLOOR_CELL_SIZE } from "./gridart.ts";

export class Floater {
  xy: Point;
  velocity: Point;
  z: number;
  velZ: number;
  frame: number;
  spin: number;
  collected: boolean;

  #callbacks: BreakableBlockPickupCallbacks;

  constructor(xy: Point, z: number, callbacks: BreakableBlockPickupCallbacks) {
    this.xy = xy;
    this.velocity = new Point(0, 0);

    this.z = z;
    this.velZ = 0;

    this.frame = 0;
    this.spin = 0;

    this.collected = false;

    this.#callbacks = callbacks;
  }

  update() {
    let bbox = this.bbox;
    let { displacement, dxy } = displace(
      bbox,
      this.velocity,
      (r) => getHuntMode().getContact(r),
      { bounce: 0.6 },
    );

    this.xy = this.xy.offset(displacement);
    this.velocity = dxy;

    this.velocity = new Point(this.velocity.x * 0.99, this.velocity.y * 0.99);

    this.velZ -= 0.04;
    this.z += this.velZ;
    if (this.z < 0) {
      this.z = 0;
      this.velZ = -this.velZ * 0.8; // minor bounce
    }

    this.frame += 1;
    this.spin += this.velocity.distance(new Point(0, 0)) * 4;

    let playerPos = getHuntMode().floatingPlayer;
    let dist = playerPos.distance(this.xy);
    let dir = Math.atan2(this.xy.y - playerPos.y, this.xy.x - playerPos.x);

    if (this.frame >= 60) {
      if (dist < 0.5) {
        this.collect();
      }
      let gravityAmt = (0.6 - dist) / 0.6;
      gravityAmt = Math.pow(gravityAmt, 0.8);
      if (gravityAmt > 0) {
        let dx = -Math.cos(dir) * 0.005;
        let dy = -Math.sin(dir) * 0.005;
        this.velocity = this.velocity.offset(
          new Point(gravityAmt * dx, gravityAmt * dy),
        );
      }
    }
  }
  collect() {
    if (this.collected) {
      return;
    }
    this.collected = true;
    this.#callbacks.obtain();
  }

  draw(drawpile: DrawPile, globalOffset: Point) {
    // TODO: Use some kind of global projection type
    let xyLow = this.xy
      .offset(new Point(-0.5, -0.5))
      .scale(FLOOR_CELL_SIZE)
      .offset(globalOffset.negate());
    let z = this.z / 100.0;
    let xy = new Point(
      xyLow.x,
      xyLow.y - z * 16, // not perspectivally accurate, but convincing
    );

    drawpile.add(0, () => {
      this.drawParticle(xyLow.offset(new Point(0, 2)), true);
    });
    if (this.frame >= 1200) {
      if (this.frame % 30 < 15) {
        return;
      }
    } else if (this.frame >= 960) {
      if (this.frame % 60 < 30) {
        return;
      }
    } else if (this.frame >= 720) {
      if (this.frame % 120 < 60) {
        return;
      }
    }
    drawpile.add(128, () => {
      this.drawParticle(xy, false);
    });
  }

  get alive(): boolean {
    return !this.collected && this.frame < 1440;
  }

  get bbox(): Circle {
    let w = 0.25;
    let h = 0.25;
    return new Circle(this.xy, w / 2);
  }
  drawParticle(projected: Point, isShadow: boolean): any {
    this.#callbacks.drawParticle(
      projected,
      isShadow,
      ((0 * Math.PI) / 4) * Math.cos(this.spin),
    );
  }
}
